package builtins

import (
	"context"
	"errors"
	"ucl.lmika.dev/ucl"
)

func Lists() ucl.Module {
	return ucl.Module{
		Name: "lists",
		Builtins: map[string]ucl.BuiltinHandler{
			"first": listFirst,
			"uniq":  listUniq,
		},
	}
}

func listFirst(ctx context.Context, args ucl.CallArgs) (any, error) {
	var (
		what  ucl.Object
		count int
	)

	if err := args.Bind(&what, &count); err != nil {
		return nil, err
	}

	if count == 0 {
		return ucl.NewListObject(), nil
	}

	newList := ucl.NewListObject()

	switch t := what.(type) {
	case ucl.Listable:
		if count < 0 {
			count = t.Len() + count
		}

		for i := 0; i < min(count, t.Len()); i++ {
			newList.Append(t.Index(i))
		}
	case ucl.Iterable:
		if count < 0 {
			return nil, errors.New("negative counts not supported on iters")
		}

		for i := 0; t.HasNext() && i < count; i++ {
			v, err := t.Next(ctx)
			if err != nil {
				return nil, err
			}
			newList.Append(v)
		}
	default:
		return nil, errors.New("expected listable")
	}

	return newList, nil
}

func eachListOrIterItem(ctx context.Context, o ucl.Object, f func(int, ucl.Object) error) error {
	switch t := o.(type) {
	case ucl.Listable:
		for i := 0; i < t.Len(); i++ {
			if err := f(i, t.Index(i)); err != nil {
				return err
			}
		}
		return nil
	case ucl.Iterable:
		idx := 0
		for t.HasNext() {
			v, err := t.Next(ctx)
			if err != nil {
				return err
			}
			if err := f(idx, v); err != nil {
				return err
			}
			idx++
		}
	}
	return errors.New("expected listable")
}

type uniqKey struct {
	sVal string
	iVal int
}

func listUniq(ctx context.Context, args ucl.CallArgs) (any, error) {
	var (
		what ucl.Object
	)

	if err := args.Bind(&what); err != nil {
		return nil, err
	}

	seen := make(map[uniqKey]bool)
	found := ucl.NewListObject()

	if err := eachListOrIterItem(ctx, what, func(idx int, v ucl.Object) error {
		var key uniqKey
		switch v := v.(type) {
		case ucl.StringObject:
			key = uniqKey{sVal: string(v)}
		case ucl.IntObject:
			key = uniqKey{iVal: int(v)}
		default:
			return errors.New("expected string or int")
		}

		if !seen[key] {
			seen[key] = true
			found.Append(v)
		}

		return nil
	}); err != nil {
		return nil, err
	}

	return found, nil
}
